using OpenTK.Graphics.OpenGL4;
using OpenTK.Mathematics;
using OpenTK.Windowing.Common;
using OpenTK.Windowing.Desktop;
using OpenTK.Windowing.GraphicsLibraryFramework;
// https://docs.sixlabors.com/api/ImageSharp/SixLabors.ImageSharp.Image.html
using SixLabors.ImageSharp;
using SixLabors.ImageSharp.Metadata.Profiles.Exif;
using System;
using System.Runtime.CompilerServices;
using Image = SixLabors.ImageSharp.Image;

namespace SemiColinGames;

public class CameraInfo {
  public readonly Vector2i Resolution;

  private CameraInfo(Vector2i resolution) {
    Resolution = resolution;
  }

  public static readonly CameraInfo NIKON_D7000 = new CameraInfo(new Vector2i(4928, 3264));
  public static readonly CameraInfo IPHONE_12_MINI = new CameraInfo(new Vector2i(4032, 3024));
}

public class Shader : IDisposable {
  int Handle;

  public Shader() {
    int VertexShader;
    int FragmentShader;

    string VertexShaderSource = @"
#version 330

layout(location = 0) in vec3 aPosition;
layout(location = 1) in vec2 aTexCoord;

out vec2 texCoord;

uniform mat4 projection;

void main(void) {
    texCoord = aTexCoord;
    gl_Position = vec4(aPosition, 1.0) * projection;
}";

    string FragmentShaderSource = @"
#version 330

out vec4 outputColor;
in vec2 texCoord;
uniform sampler2D texture0;
uniform vec4 color;

void main() {
    outputColor = texture(texture0, texCoord) * color;
}";

    VertexShader = GL.CreateShader(ShaderType.VertexShader);
    GL.ShaderSource(VertexShader, VertexShaderSource);

    FragmentShader = GL.CreateShader(ShaderType.FragmentShader);
    GL.ShaderSource(FragmentShader, FragmentShaderSource);

    GL.CompileShader(VertexShader);

    int success;
    GL.GetShader(VertexShader, ShaderParameter.CompileStatus, out success);
    if (success == 0) {
      string infoLog = GL.GetShaderInfoLog(VertexShader);
      Console.WriteLine(infoLog);
    }

    GL.CompileShader(FragmentShader);

    GL.GetShader(FragmentShader, ShaderParameter.CompileStatus, out success);
    if (success == 0) {
      string infoLog = GL.GetShaderInfoLog(FragmentShader);
      Console.WriteLine(infoLog);
    }
    Handle = GL.CreateProgram();

    GL.AttachShader(Handle, VertexShader);
    GL.AttachShader(Handle, FragmentShader);

    GL.LinkProgram(Handle);

    GL.GetProgram(Handle, GetProgramParameterName.LinkStatus, out success);
    if (success == 0) {
      string infoLog = GL.GetProgramInfoLog(Handle);
      Console.WriteLine(infoLog);
    }

    GL.DetachShader(Handle, VertexShader);
    GL.DetachShader(Handle, FragmentShader);
    GL.DeleteShader(FragmentShader);
    GL.DeleteShader(VertexShader);
  }

  public void Use() {
    GL.UseProgram(Handle);
  }

  private bool disposedValue = false;

  protected virtual void Dispose(bool disposing) {
    if (!disposedValue) {
      GL.DeleteProgram(Handle);
      disposedValue = true;
    }
  }

  ~Shader() {
    if (disposedValue == false) {
      Console.WriteLine("~Shader(): resource leak? Dispose() should be called manually.");
    }
  }

  public void Dispose() {
    Dispose(true);
    GC.SuppressFinalize(this);
  }

  public int GetAttribLocation(string name) {
    return GL.GetAttribLocation(Handle, name);
  }

  public int GetUniformLocation(string name) {
    return GL.GetUniformLocation(Handle, name);
  }

}

public class Texture : IDisposable {
  public int Handle;
  // FIXME: use a Vector2i for dimensions.
  public int Width;
  public int Height;

  public Texture(Image<Rgba32> image) {
    Width = image.Width;
    Height = image.Height;
    Console.WriteLine($"image loaded: {Width}x{Height}");

    //foreach (IExifValue exif in image.Metadata.ExifProfile.Values) {
    //  Console.WriteLine($"{exif.Tag} : {exif.GetValue()}");
    //}
    byte[] pixelBytes = new byte[Width * Height * Unsafe.SizeOf<Rgba32>()];
    image.CopyPixelDataTo(pixelBytes);
    image.Dispose();

    Handle = GL.GenTexture();
    GL.ActiveTexture(TextureUnit.Texture0);
    GL.BindTexture(TextureTarget.Texture2D, Handle);
    GL.TexImage2D(TextureTarget.Texture2D, 0, PixelInternalFormat.Rgba, image.Width, image.Height, 0, PixelFormat.Rgba, PixelType.UnsignedByte, pixelBytes);
    GL.TexParameter(TextureTarget.Texture2D, TextureParameterName.TextureMinFilter, (int) TextureMinFilter.LinearMipmapLinear); // FIXME: is this right?
    GL.TexParameter(TextureTarget.Texture2D, TextureParameterName.TextureMagFilter, (int) TextureMagFilter.Nearest);
    GL.TexParameter(TextureTarget.Texture2D, TextureParameterName.TextureWrapS, (int) TextureWrapMode.ClampToBorder);
    GL.TexParameter(TextureTarget.Texture2D, TextureParameterName.TextureWrapT, (int) TextureWrapMode.ClampToBorder);
    float[] borderColor = { 0.0f, 0.0f, 0.0f, 1.0f };
    GL.TexParameter(TextureTarget.Texture2D, TextureParameterName.TextureBorderColor, borderColor);
    GL.GenerateMipmap(GenerateMipmapTarget.Texture2D);
  }

  private bool disposedValue = false;

  protected virtual void Dispose(bool disposing) {
    if (!disposedValue) {
      GL.DeleteTexture(Handle);
      disposedValue = true;
    }
  }

  ~Texture() {
    if (!disposedValue) {
      Console.WriteLine("~Texture(): resource leak? Dispose() should be called manually.");
    }
  }

  public void Dispose() {
    Dispose(true);
    GC.SuppressFinalize(this);
  }
}

public class UiGeometry {
  public static Vector2i MIN_WINDOW_SIZE = new Vector2i(640, 480);
  private static CameraInfo activeCamera = CameraInfo.NIKON_D7000;

  public readonly Vector2i WindowSize;
  public readonly List<Box2i> ThumbnailBoxes = new();
  public readonly Box2i PhotoBox;

  public UiGeometry() : this(MIN_WINDOW_SIZE) {}

  public UiGeometry(Vector2i windowSize) {
    WindowSize = windowSize;

    int thumbnailHeight = 150;
    int thumbnailWidth = (int) 1.0 * thumbnailHeight * activeCamera.Resolution.X / activeCamera.Resolution.Y;
    int numBoxes = 10; // FIXME
    for (int i = 0; i < numBoxes; i++) {
      Box2i box = Util.makeBox(WindowSize.X - thumbnailWidth, i * thumbnailHeight, thumbnailWidth, thumbnailHeight);
      ThumbnailBoxes.Add(box);
    }

    PhotoBox = new(0, 0, WindowSize.X - thumbnailWidth, WindowSize.Y);
  }
}

public static class Util {
  public static Box2i makeBox(int left, int top, int width, int height) {
    return new Box2i(left, top, left + width, top + height);
  }
}

public class Game : GameWindow {
  public Game(GameWindowSettings gwSettings, NativeWindowSettings nwSettings) : base(gwSettings, nwSettings) { }

  static Texture TEXTURE_WHITE;

  UiGeometry geometry = new();

  // Four points, each consisting of (x, y, z, tex_x, tex_y).
  float[] vertices = new float[20];

  // Indices to draw a rectangle from two triangles.
  uint[] indices = {
    0, 1, 3,   // first triangle
    1, 2, 3    // second triangle
  };

  int VertexBufferObject;
  int ElementBufferObject;
  int VertexArrayObject;
  List<Texture> textures = new();
  int textureIndex = 0;
  Shader shader;
  Matrix4 projection;

  protected override void OnUpdateFrame(FrameEventArgs e) {
    base.OnUpdateFrame(e);

    KeyboardState input = KeyboardState;

    if (input.IsKeyDown(Keys.Escape)) {
      Close();
    }

    if (input.IsKeyPressed(Keys.Down)) {
      if (textureIndex < textures.Count - 1) {
        textureIndex++;
      }
    }
    if (input.IsKeyPressed(Keys.Up)) {
      if (textureIndex > 0) {
        textureIndex--;
      }
    }
  }

  protected override void OnLoad() {
    base.OnLoad();

    GL.ClearColor(0f, 0f, 0f, 1f);

    VertexArrayObject = GL.GenVertexArray();
    GL.BindVertexArray(VertexArrayObject);

    VertexBufferObject = GL.GenBuffer();
    ElementBufferObject = GL.GenBuffer();

    GL.BindBuffer(BufferTarget.ArrayBuffer, VertexBufferObject);
    GL.BufferData(BufferTarget.ArrayBuffer, vertices.Length * sizeof(float), vertices, BufferUsageHint.DynamicDraw);
    GL.BindBuffer(BufferTarget.ElementArrayBuffer, ElementBufferObject);
    GL.BufferData(BufferTarget.ElementArrayBuffer, indices.Length * sizeof(uint), indices, BufferUsageHint.DynamicDraw);

    shader = new Shader();
    shader.Use();

    // Because there's 5 floats between the start of the first vertex and the start of the second,
    // the stride is 5 * sizeof(float).
    // This will now pass the new vertex array to the buffer.
    var vertexLocation = shader.GetAttribLocation("aPosition");
    GL.EnableVertexAttribArray(vertexLocation);
    GL.VertexAttribPointer(vertexLocation, 3, VertexAttribPointerType.Float, false, 5 * sizeof(float), 0);

    // Next, we also setup texture coordinates. It works in much the same way.
    // We add an offset of 3, since the texture coordinates comes after the position data.
    // We also change the amount of data to 2 because there's only 2 floats for texture coordinates.
    var texCoordLocation = shader.GetAttribLocation("aTexCoord");
    GL.EnableVertexAttribArray(texCoordLocation);
    GL.VertexAttribPointer(texCoordLocation, 2, VertexAttribPointerType.Float, false, 5 * sizeof(float), 3 * sizeof(float));

    // Create a blank white texture.
    Image<Rgba32> white1x1 = new(1, 1, new Rgba32(255, 255, 255));
    TEXTURE_WHITE = new Texture(white1x1);

    // Load textures from JPEGs.
    string[] files = Directory.GetFiles(@"c:\users\colin\pictures\photos\2023\06\27\");
    foreach (string file in files) {
      if (file.ToLower().EndsWith(".jpg")) {
        Image<Rgba32> image = Image.Load<Rgba32>(file);
        textures.Add(new Texture(image));
      }
    }
  }

  protected override void OnUnload() {
    base.OnUnload();
  }

  protected override void OnRenderFrame(FrameEventArgs e) {
    base.OnRenderFrame(e);
    GL.Clear(ClearBufferMask.ColorBufferBit);
    GL.BindBuffer(BufferTarget.ArrayBuffer, VertexBufferObject);
    GL.ActiveTexture(TextureUnit.Texture0);

    Texture active = textures[textureIndex];

    // FIXME: make a function for scaling & centering one box on another.
    double scaleX = 1.0 * geometry.PhotoBox.Size.X / active.Width;
    double scaleY = 1.0 * geometry.PhotoBox.Size.Y / active.Height;
    double scale = Math.Min(scaleX, scaleY);
    int renderWidth = (int) (active.Width * scale);
    int renderHeight = (int) (active.Height * scale);
    Box2i photoBox = Util.makeBox((int) geometry.PhotoBox.Center.X - renderWidth / 2, (int) geometry.PhotoBox.Center.Y - renderHeight / 2, renderWidth, renderHeight);
    DrawTexture(active, photoBox); 
    for (int i = 0; i < textures.Count; i++) {
      Box2i box = geometry.ThumbnailBoxes[i];
      DrawTexture(textures[i], box);
      if (i == textureIndex) {
        DrawBox(box, 5, Color4.Black);
        DrawBox(box, 3, Color4.White);
      }
    }

    SwapBuffers();
  }

  void DrawTexture(Texture texture, Box2i box) {
    DrawTexture(texture, box, Color4.White);
  }

  void DrawTexture(Texture texture, Box2i box, Color4 color) {
    GL.Uniform4(shader.GetUniformLocation("color"), color);
    SetVertices(box.Min.X, box.Min.Y, box.Size.X, box.Size.Y);
    GL.BufferData(BufferTarget.ArrayBuffer, vertices.Length * sizeof(float), vertices, BufferUsageHint.DynamicDraw);
    GL.BindTexture(TextureTarget.Texture2D, texture.Handle);
    GL.DrawElements(PrimitiveType.Triangles, indices.Length, DrawElementsType.UnsignedInt, 0);
  }

  void DrawBox(Box2i box, int thickness, Color4 color) {
    DrawTexture(TEXTURE_WHITE, Util.makeBox(box.Min.X, box.Min.Y, box.Size.X, thickness), color);
    DrawTexture(TEXTURE_WHITE, Util.makeBox(box.Min.X, box.Min.Y, thickness, box.Size.Y), color);
    DrawTexture(TEXTURE_WHITE, Util.makeBox(box.Min.X, box.Max.Y - thickness, box.Size.X, thickness), color);
    DrawTexture(TEXTURE_WHITE, Util.makeBox(box.Max.X - thickness, box.Min.Y, thickness, box.Size.Y), color);
  }

  protected override void OnResize(ResizeEventArgs e) {
    base.OnResize(e);
    Console.WriteLine($"OnResize: {e.Width}x{e.Height}");

    geometry = new UiGeometry(e.Size);

    projection = Matrix4.CreateOrthographicOffCenter(0f, e.Width, e.Height, 0f, -1f, 1f);
    GL.UniformMatrix4(shader.GetUniformLocation("projection"), true, ref projection);
    GL.Viewport(0, 0, e.Width, e.Height);
  }

  private void SetVertices(float left, float top, float width, float height) {
    // top left
    vertices[0] = left;
    vertices[1] = top;
    vertices[2] = 0f;
    vertices[3] = 0f;
    vertices[4] = 0f;

    // top right
    vertices[5] = left + width;
    vertices[6] = top;
    vertices[7] = 0f;
    vertices[8] = 1f;
    vertices[9] = 0f;

    // bottom right
    vertices[10] = left + width;
    vertices[11] = top + height;
    vertices[12] = 0f;
    vertices[13] = 1f;
    vertices[14] = 1f;

    // bottom left
    vertices[15] = left;
    vertices[16] = top + height;
    vertices[17] = 0f;
    vertices[18] = 0f;
    vertices[19] = 1f;
  }
}

static class Program {
  static void Main(string[] args) {
    List<MonitorInfo> monitors = Monitors.GetMonitors();
    MonitorInfo bestMonitor = monitors[0];
    int bestResolution = bestMonitor.HorizontalResolution * bestMonitor.VerticalResolution;
    for (int i = 1; i < monitors.Count; i++) {
      MonitorInfo monitor = monitors[i];
      int resolution = monitor.HorizontalResolution * monitor.VerticalResolution;
      if (resolution > bestResolution) {
        bestResolution = resolution;
        bestMonitor = monitor;
      }
    }
    Console.WriteLine($"best monitor: {bestMonitor.HorizontalResolution}x{bestMonitor.VerticalResolution}");
    GameWindowSettings gwSettings = new();
    gwSettings.RenderFrequency = 60.0;

    NativeWindowSettings nwSettings = new();
    nwSettings.WindowState = WindowState.Normal;
    nwSettings.CurrentMonitor = bestMonitor.Handle;
    nwSettings.Location = new Vector2i(bestMonitor.WorkArea.Min.X + 1, bestMonitor.WorkArea.Min.Y + 31);
    nwSettings.Size = new Vector2i(bestMonitor.WorkArea.Size.X - 2, bestMonitor.WorkArea.Size.Y - 32);
    nwSettings.MinimumSize = UiGeometry.MIN_WINDOW_SIZE;
    nwSettings.Title = "Totte";
    // FIXME: nwSettings.Icon = ...

    using (Game game = new Game(gwSettings, nwSettings)) {
      game.Run();
    }
  }
}

